PROMPT_KNOWLEDGE = """\
## Use the following references from the knowledge base if it helps:

<references>
{references}
<references>
"""

PROMPT_CONTEXT = """\
## Use the following context if it helps:

<context>
{context}
</context>
"""

PROMPT_PREVIOUS_INTERACTIONS = """\
## Answer the question considering the previous interactions:

<previous_interactions>
{previous_interactions}
</previous_interactions>
"""

PROMPT_JSON_SCHEMA = """\
## Provide your output using the following JSON schema:

<json_schema>
{json_schema}
</json_schema>
"""



import hashlib
import re
from typing import List

from pydantic import BaseModel
from tqdm import tqdm

from gwenflow.logger import logger
from gwenflow.types.document import Document

try:
    import tiktoken
except ImportError as e:
    raise ImportError("`tiktoken` is not installed. Please install it with `pip install tiktoken`.") from e


class TokenTextSplitter(BaseModel):
    chunk_size: int = 500
    chunk_overlap: int = 100
    encoding_name: str = "cl100k_base"
    strip_whitespace: bool = False
    normalize_text: bool = False

    def split_text(self, text: str) -> List[str]:
        _tokenizer = tiktoken.get_encoding(self.encoding_name)
        input_ids = _tokenizer.encode(text)

        if self.normalize_text:
            text = text.replace("\n", " ").replace("\r", " ")
            text = re.sub(" +", " ", text)

        splits: List[str] = []
        start_idx = 0
        cur_idx = min(start_idx + self.chunk_size, len(input_ids))
        chunk_ids = input_ids[start_idx:cur_idx]
        while start_idx < len(input_ids):
            splits.append(_tokenizer.decode(chunk_ids))
            if cur_idx == len(input_ids):
                break
            start_idx += self.chunk_size - self.chunk_overlap
            cur_idx = min(start_idx + self.chunk_size, len(input_ids))
            chunk_ids = input_ids[start_idx:cur_idx]

        if self.strip_whitespace:
            splits = [s.strip() for s in splits]

        return splits

    def split_documents(
        self, documents: list[Document], chunk_fields: list = None, metadata_fields: list = None
    ) -> list[Document]:
        if metadata_fields is None:
            metadata_fields = []
        if chunk_fields is None:
            chunk_fields = []
        chunks = []
        for document in tqdm(documents):
            if not document.id:
                logger.warning(f"Missing id on document: {document['content']}. Skipping document.")
                continue

            # content
            content = ""
            if document.content:
                content = document.content
            elif chunk_fields:
                content = [f"{f.upper()}: {document.metadata.get(f)}" for f in chunk_fields if document.metadata.get(f)]
                content = ", ".join(content)
            else:
                content = [f"{k.upper()}: {v}" for k, v in document.metadata.items() if v]
                content = ", ".join(content)

            if not content:
                logger.warning(f"Missing content on document id: {document.id}. Skipping document.")
                continue

            # meta
            metadata = document.metadata
            if metadata_fields:
                metadata = {}
                for f in metadata_fields:
                    metadata[f] = document.metadata.get(f)
            metadata["document_id"] = document.id  # keep original doc id

            # split
            splitted_text = self.split_text(text=content)
            for i, chunk in enumerate(splitted_text):
                metadata["chunk_id"] = f"chunk_{i}"
                _id = hashlib.md5("-".join([document.id, str(i)]).encode(), usedforsecurity=False).hexdigest()
                chunks.append(Document(id=_id, content=chunk, metadata=metadata))

        return chunks



from gwenflow.types.message import Message
from gwenflow.types.document import Document
from gwenflow.types.usage import Usage
from gwenflow.types.response import Function, ToolCall, ToolResponse, ModelResponse, AgentResponse
from gwenflow.types.items import ItemHelpers

__all__ = [
    "Message",
    "Document",
    "Usage",
    "Function",
    "ToolCall",
    "ToolResponse",
    "ModelResponse",
    "AgentResponse",
    "ItemHelpers",
]


from typing import Any

from gwenflow.tools.base import BaseTool
from gwenflow.tools.mcp.server import MCPServerSse, MCPServerSseParams


class MCPTool(BaseTool):
    def _run(self, **kwargs: Any) -> Any:
        params = MCPServerSseParams(url=self.url, headers=self.headers)
        server = MCPServerSse(params=params)
        return server.call_tool(tool=self.tool, **kwargs)

    @classmethod
    async def from_server(cls, url: str, name: str, headers: dict = None) -> "MCPTool":
        params = MCPServerSseParams(url=url, headers=headers)
        server = MCPServerSse(params=params)
        await server.connect()

        tool = None
        response = await server.list_tools()
        for t in response.tools:
            if t.name == name:
                tool = t
                break

        if not tool:
            raise ValueError("Tool not found not in MCP.")

        return MCPTool(
            name=tool.__name__,
            description=tool.description,
            params_json_schema=tool.inputSchema,
            tool_type="mcp",
        )



import shutil
import subprocess


def validate_docker_installation(agent_role: str) -> None:
    """Check if Docker is installed and running.

    Args:
        agent_role (str): The role of the agent requesting validation.

    Raises:
        RuntimeError: If Docker is missing or not running.
    """
    if not shutil.which("docker"):
        raise RuntimeError(
            f"Docker is not installed. Please install Docker to use code "
            f"execution with agent: {agent_role}"
        )

    try:
        subprocess.run(
            ["docker", "info"],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Docker is not running. Please start Docker to use code "
            f"execution with agent: {agent_role}"
        ) from e



from typing import List, Literal, Optional

from pydantic import BaseModel


class ResponseBase(BaseModel):
    sequence_number: int

class Logprob(BaseModel): # TODO move this later 
    token: str
    logprob: float
    bytes: Optional[List[int]] = None

class ReasoningItem(BaseModel): # TODO move this later
    effort: Literal['low', 'medium', 'high']
    summary: Literal['auto', 'concise', 'detailed']



import copy
from typing import Dict, List, Union

from .message import Message


class ItemHelpers:
    @classmethod
    def input_to_message_list(
        cls,
        input: Union[str, List[Message], List[Dict[str, str]]],
    ) -> List[Message]:
        """Converts a string or list of messages into a list of messages."""
        if isinstance(input, str):
            return [Message(role="user", content=input)]
        messages = copy.deepcopy(input)
        for i, message in enumerate(messages):
            if isinstance(message, dict):
                messages[i] = Message(**message)
        return messages



class GwenflowException(Exception):  # noqa: N818
    """Base class for all exceptions in Gwenflow."""


class MaxTurnsExceeded(GwenflowException):
    """Exception raised when the maximum number of turns is exceeded."""

    message: str

    def __init__(self, message: str):
        self.message = message


class ModelBehaviorError(GwenflowException):
    """Exception raised when the model does something unexpected."""

    message: str

    def __init__(self, message: str):
        self.message = message


class UserError(GwenflowException):
    """Exception raised when the user makes an error."""

    message: str

    def __init__(self, message: str):
        self.message = message



import asyncio
import json
import re
import uuid
from typing import Any, AsyncIterator, Dict, Iterator, List, Literal, Optional, Union

from pydantic import UUID4, BaseModel, ConfigDict, Field, field_validator, model_validator

from gwenflow.agents.prompts import PROMPT_CONTEXT, PROMPT_JSON_SCHEMA, PROMPT_KNOWLEDGE
from gwenflow.llms import ChatBase, ChatOpenAI
from gwenflow.logger import logger
from gwenflow.memory import ChatMemoryBuffer
from gwenflow.retriever import Retriever
from gwenflow.telemetry import tracer
from gwenflow.tools import BaseTool
from gwenflow.tools.mcp import MCPServer, MCPUtil
from gwenflow.types import AgentResponse, ItemHelpers, Message, ToolCall, ToolResponse

DEFAULT_MAX_TURNS = 10


class Agent(BaseModel):
    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    """The unique id of the agent."""

    name: str
    """The name of the agent."""

    description: str | None = None
    """A description of the agent, used as a handoff, so that the manager knows what it does."""

    system_prompt: str | None = None
    """"System prompt"""

    instructions: str | List[str] | None = None
    """The instructions for the agent."""

    response_model: Dict | None = None
    """Response model."""

    llm: Optional[ChatBase] = Field(None, validate_default=True)
    """The model implementation to use when invoking the LLM."""

    tools: List[BaseTool] = Field(default_factory=list)
    """A list of tools that the agent can use."""

    mcp_servers: List[MCPServer] = Field(default_factory=list)
    """A list of MCP servers that the agent can use."""

    tool_choice: Literal["auto", "required", "none"] | str | None = None
    """The tool choice to use when calling the model."""

    reasoning_model: Optional[ChatBase] = Field(None, validate_default=True)
    """Reasoning model."""

    history: ChatMemoryBuffer | None = None
    """Historcal messages for the agent."""

    retriever: Optional[Retriever] = None
    """Retriever for the agent."""

    team: List["Agent"] | None = None
    """Team of agents."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @field_validator("id", mode="before")
    @classmethod
    def deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise ValueError("This field is not to be set by the user.")

    @field_validator("llm", mode="before")
    @classmethod
    def set_llm(cls, v: Optional[Any]) -> Any:
        llm = v or ChatOpenAI(model="gpt-4o-mini")
        return llm

    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if self.llm:
            if self.history is None:
                token_limit = self.llm.get_context_window_size()
                self.history = ChatMemoryBuffer(token_limit=token_limit)
            if self.response_model:
                self.llm.response_format = {"type": "json_object"}
            if self.tools or self.mcp_servers:
                self.llm.tools = self.get_all_tools()
                self.llm.tool_choice = self.tool_choice
            else:
                self.llm.tools = None
                self.llm.tool_choice = None
        return self

    def _format_context(self, context: Optional[Union[str, Dict[str, str]]]) -> str:
        text = ""
        if isinstance(context, str):
            text = f"<context>\n{context}\n</context>\n\n"
        elif isinstance(context, dict):
            for key in context.keys():
                text += f"<{key}>\n"
                text += context.get(key) + "\n"
                text += f"</{key}>\n\n"
        return text

    def get_system_prompt(
        self,
        task: str,
        context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> str:
        """Get the system prompt for the agent."""
        if self.system_prompt:
            return self.system_prompt

        prompt = "Your name is {name}.".format(name=self.name)

        if self.instructions:
            if isinstance(self.instructions, str):
                prompt += " {instructions}".format(instructions=self.instructions)
            elif isinstance(self.instructions, list):
                instructions = "\n".join([f"- {i}" for i in self.instructions])
                prompt += "\n\n## Instructions:\n{instructions}".format(instructions=instructions)

        prompt += "\n\n"

        if self.response_model:
            prompt += PROMPT_JSON_SCHEMA.format(json_schema=json.dumps(self.response_model, indent=4)).strip()
            prompt += "\n\n"

        if self.retriever:
            references = self.retriever.search(query=task)
            if len(references) > 0:
                references = [r.content for r in references]
                prompt += PROMPT_KNOWLEDGE.format(references="\n\n".join(references)).strip()
                prompt += "\n\n"

        if context is not None:
            prompt += PROMPT_CONTEXT.format(context=self._format_context(context)).strip()
            prompt += "\n\n"

        return prompt.strip()

    def reason(
        self,
        input: Union[str, List[Message], List[Dict[str, str]]],
    ) -> AgentResponse:
        if self.reasoning_model is None:
            return None

        logger.debug("Reasoning...")

        reasoning_agent = Agent(
            name="ReasoningAgent",
            instructions=[
                "You are a meticulous and thoughtful assistant that solves a problem by thinking through it step-by-step.",
                "Carefully analyze the task by spelling it out loud.",
                "Then break down the problem by thinking through it step by step and develop multiple strategies to solve the problem."
                "Work through your plan step-by-step, executing any tools as needed for each step.",
                "Do not call any tool or try to solve the problem yourself.",
                "Your task is to provide a plan step-by-step, not to solve the problem yourself.",
            ],
            llm=self.reasoning_model,
            tools=self.tools,
        )

        response = reasoning_agent.run(input)

        # only keep text outside <think>
        reasoning_content = re.sub(r'<think>.*?</think>', '', response.content, flags=re.DOTALL)
        reasoning_content = reasoning_content.strip()
        if not reasoning_content:
            return None

        self.history.add_message(
            Message(
                role="assistant",
                content=f"I have worked through this problem in-depth and my reasoning is summarized below.\n\n{reasoning_content}",
            )
        )

        logger.debug("Thought:\n" + reasoning_content)

        return response

    async def areason(
        self,
        input: Union[str, List[Message], List[Dict[str, str]]],
    ) -> AgentResponse:
        if self.reasoning_model is None:
            return None

        logger.debug("Reasoning...")

        reasoning_agent = Agent(
            name="ReasoningAgent",
            instructions=[
                "You are a meticulous and thoughtful assistant that solves a problem by thinking through it step-by-step.",
                "Carefully analyze the task by spelling it out loud.",
                "Then break down the problem by thinking through it step by step and develop multiple strategies to solve the problem."
                "Work through your plan step-by-step, executing any tools as needed for each step.",
                "Do not call any tool or try to solve the problem yourself.",
                "Your task is to provide a plan step-by-step, not to solve the problem yourself.",
            ],
            llm=self.reasoning_model,
            tools=self.tools,
        )

        response = await reasoning_agent.run(input)

        # only keep text outside <think>
        reasoning_content = re.sub(r'<think>.*?</think>', '', response.content, flags=re.DOTALL)
        reasoning_content = reasoning_content.strip()
        if not reasoning_content:
            return None

        self.history.add_message(
            Message(
                role="assistant",
                content=f"I have worked through this problem in-depth and my reasoning is summarized below.\n\n{reasoning_content}",
            )
        )

        logger.debug("Thought:\n" + reasoning_content)

        return response

    def get_all_tools(self) -> list[BaseTool]:
        """All agent tools, including MCP tools and function tools."""
        tools = self.tools
        if self.mcp_servers:
            mcp_tools = asyncio.run(MCPUtil.get_all_function_tools(self.mcp_servers))
            # tools += MCPUtil.get_all_function_tools(self.mcp_servers)
            tools += mcp_tools
        return tools

    @tracer.tool(name="Tool Call")
    def run_tool(self, tool_call: ToolCall) -> Message:

        tool_execution = ToolResponse(
            tool_call_id=tool_call.id,
            tool_name=tool_call.function.name,
        )

        tool_map  = {tool.name: tool for tool in self.get_all_tools()}

        if tool_call.function.name not in tool_map.keys():
            logger.error(f"Tool {tool_call.function} does not exist")
            tool_execution.result = f"Tool {tool_call.function} does not exist"
            return tool_execution.to_message()

        try:
            tool = tool_map[tool_call.function.name]
            arguments = json.loads(tool_call.function.arguments)
            logger.debug(f"Tool call: {tool_call.function}({arguments})")
            tool_execution.result = tool.run(**arguments)
            if tool_execution.result:
                return tool_execution.to_message()

        except Exception as e:
            logger.error(f"Error executing tool '{tool_call.function}': {e}")

        tool_execution.result = f"Error executing tool '{tool_call.function}'"
        return tool_execution.to_message()

    def execute_tool_calls(self, tool_calls: List[ToolCall]) -> List[Message]:
        results = []
        for tool_call in tool_calls:
            result = self.run_tool(tool_call)
            if result:
                results.append(result)

        return results

    async def aexecute_tool_calls(self, tool_calls: List[ToolCall]) -> List[Message]:
        tasks = []
        for tool_call in tool_calls:
            task = asyncio.create_task(asyncio.to_thread(self.run_tool, tool_call))
            tasks.append(task)

        results = await asyncio.gather(*tasks)

        return results

    @tracer.agent(name="Agent Run")
    def run(
            self,
            input: Union[str, List[Message], List[Dict[str, str]]],
            context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> AgentResponse:

        # prepare messages and task
        messages = ItemHelpers.input_to_message_list(input)
        task = messages[-1].content

        # init agent response
        agent_response = AgentResponse()

        # history
        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        # add reasoning
        if self.reasoning_model:
            messages_for_reasoning_model = [m.to_dict() for m in self.history.get()]
            response = self.reason(messages_for_reasoning_model)
            agent_response.reasoning_content = response.reasoning_content
            agent_response.usage.add(response.usage)

        num_turns_available = DEFAULT_MAX_TURNS

        while num_turns_available > 0:

            num_turns_available -= 1

            # format messages
            messages_for_model = [m.to_dict() for m in self.history.get()]

            # call llm and tool
            response = self.llm.invoke(input=messages_for_model)

            # usage
            agent_response.usage.add(response.usage)

            # keep answer in memory
            tool_calls = [t.model_dump() for t in response.tool_calls]
            _message = Message(role=response.role, content=response.content, tool_calls=tool_calls)
            self.history.add_message(_message.model_dump())

            # stop if not tool call
            if not response.tool_calls:
                agent_response.content = response.content
                agent_response.messages.append(_message)
                break

            # handle tool calls
            if response.tool_calls and self.get_all_tools():
                tool_messages = self.execute_tool_calls(tool_calls=response.tool_calls)
                for m in tool_messages:
                    self.history.add_message(m)
                    agent_response.messages.append(m)

            if self.tool_choice == "required":
                self.tool_choice = "auto"

        # format response
        if self.response_model:
            agent_response.parsed = json.loads(agent_response.content)

        agent_response.finish_reason = "stop"

        return agent_response

    @tracer.agent(name="Agent Arun")
    async def arun(
            self,
            input: Union[str, List[Message], List[Dict[str, str]]],
            context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> AgentResponse:

        messages = ItemHelpers.input_to_message_list(input)
        task = messages[-1].content

        agent_response = AgentResponse()

        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        if self.reasoning_model:
            messages_for_reasoning_model = [m.to_dict() for m in self.history.get()]
            reasoning_agent_response = await self.areason(messages_for_reasoning_model)
            agent_response.reasoning_content = reasoning_agent_response.reasoning_content
            agent_response.usage.add(reasoning_agent_response.usage)

        num_turns_available = DEFAULT_MAX_TURNS

        while num_turns_available > 0:

            num_turns_available -= 1

            messages_for_model = [m.to_dict() for m in self.history.get()]

            response = await self.llm.ainvoke(input=messages_for_model)

            # usage
            agent_response.usage.add(response.usage)

            tool_calls = [t.model_dump() for t in response.tool_calls]
            _message = Message(role=response.role, content=response.content, tool_calls=tool_calls)
            self.history.add_message(_message.model_dump())

            if not response.tool_calls:
                agent_response.content = response.content
                agent_response.messages.append(_message)
                break

            # handle tool calls
            if response.tool_calls and self.get_all_tools():
                tool_messages = await self.aexecute_tool_calls(tool_calls=response.tool_calls)
                for m in tool_messages:
                    self.history.add_message(m)
                    agent_response.messages.append(m)

            if self.tool_choice == "required":
                self.tool_choice = "auto"

        if self.response_model:
            agent_response.content = json.loads(agent_response.content)
            agent_response.parsed = agent_response.content

        agent_response.finish_reason = "stop"

        return agent_response

    @tracer.agent(name="Agent Stream")
    def run_stream(
        self,
        input: Union[str, List[Message], List[Dict[str, str]]],
        context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> Iterator[AgentResponse]:

        messages = ItemHelpers.input_to_message_list(input)
        task = messages[-1].content

        agent_response = AgentResponse()

        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        if self.reasoning_model:
            messages_for_reasoning_model = [m.to_dict() for m in self.history.get()]
            reasoning_response = self.reason(messages_for_reasoning_model)
            agent_response.reasoning_content = reasoning_response.reasoning_content
            agent_response.usage.add(reasoning_response.usage)

        num_turns_available = DEFAULT_MAX_TURNS

        while num_turns_available > 0:
            num_turns_available -= 1

            messages_for_model = [m.to_dict() for m in self.history.get()]

            full_content = ""
            final_tool_calls: List[ToolCall] = []

            for chunk in self.llm.stream(input=messages_for_model):
                if chunk.usage:
                    agent_response.usage.add(chunk.usage)

                agent_response.content = None

                if chunk.content:
                    agent_response.content = chunk.content
                    full_content += chunk.content

                if chunk.tool_calls:
                    final_tool_calls = chunk.tool_calls

                yield agent_response

            tool_calls_dicts = [t.model_dump() for t in final_tool_calls]
            _message = Message(role="assistant", content=full_content, tool_calls=tool_calls_dicts)
            self.history.add_message(_message.model_dump())

            if not final_tool_calls:
                agent_response.content = full_content
                agent_response.messages.append(_message)
                break

            if final_tool_calls and self.get_all_tools():
                tool_messages = self.execute_tool_calls(tool_calls=final_tool_calls)
                for m in tool_messages:
                    self.history.add_message(m)
                    agent_response.messages.append(m)

            if self.tool_choice == "required":
                self.tool_choice = "auto"

        if self.response_model:
            try:
                agent_response.parsed = json.loads(full_content)
            except Exception as e:
                logger.error(f"Failed to parse JSON response: {e}")
                agent_response.parsed = {"error": f"Parse error: {str(e)}"}

        agent_response.finish_reason = "stop"

        yield agent_response

    @tracer.agent(name="Agent Astream")
    async def arun_stream(
        self,
        input: Union[str, List[Message], List[Dict[str, str]]],
        context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> AsyncIterator[AgentResponse]:

        messages = ItemHelpers.input_to_message_list(input)
        task = messages[-1].content

        agent_response = AgentResponse()

        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        if self.reasoning_model:
            messages_for_reasoning_model = [m.to_dict() for m in self.history.get()]
            reasoning_response = await self.areason(messages_for_reasoning_model)
            agent_response.reasoning_content = reasoning_response.reasoning_content
            agent_response.usage.add(reasoning_response.usage)

        num_turns_available = DEFAULT_MAX_TURNS

        while num_turns_available > 0:
            num_turns_available -= 1

            messages_for_model = [m.to_dict() for m in self.history.get()]

            full_content = ""
            final_tool_calls: List[ToolCall] = []

            async for chunk in self.llm.astream(input=messages_for_model):
                if chunk.usage:
                    agent_response.usage.add(chunk.usage)

                agent_response.content = None

                if chunk.content:
                    agent_response.content = chunk.content
                    full_content += chunk.content

                if chunk.tool_calls:
                    final_tool_calls = chunk.tool_calls

                yield agent_response

            tool_calls_dicts = [t.model_dump() for t in final_tool_calls]
            _message = Message(role="assistant", content=full_content, tool_calls=tool_calls_dicts)
            self.history.add_message(_message.model_dump())

            if not final_tool_calls:
                agent_response.content = full_content
                agent_response.messages.append(_message)
                break

            if final_tool_calls and self.get_all_tools():
                tool_messages = await self.aexecute_tool_calls(tool_calls=final_tool_calls)
                for m in tool_messages:
                    self.history.add_message(m)
                    agent_response.messages.append(m)

            if self.tool_choice == "required":
                self.tool_choice = "auto"

        if self.response_model:
            try:
                agent_response.parsed = json.loads(full_content)
            except Exception as e:
                logger.error(f"Failed to parse JSON response: {e}")
                agent_response.parsed = {"error": f"Parse error: {str(e)}"}

        agent_response.finish_reason = "stop"

        yield agent_response



import inspect
import json
from typing import Any, Callable


def safe_serialize(obj: Any) -> str:
    try:
        if hasattr(obj, "model_dump_json") and callable(obj.model_dump_json):
            return obj.model_dump_json()
        if hasattr(obj, "dict") and callable(obj.dict):
            return json.dumps(obj.dict(), default=str)
        return str(obj)
    except Exception:
        return str(obj)

def extract_user_inputs(func: Callable, args: tuple, kwargs: dict) -> str:
    try:
        sig = inspect.signature(func)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        arg_dict = dict(bound_args.arguments)

        arg_dict.pop("self", None)
        arg_dict.pop("cls", None)

        if "input" in arg_dict:
            return safe_serialize(arg_dict["input"])
        if "query" in arg_dict:
            return safe_serialize(arg_dict["query"])
        if "content" in arg_dict:
            return safe_serialize(arg_dict["content"])
        if "tool_call" in arg_dict:
            return safe_serialize(arg_dict["tool_call"])

        return json.dumps({k: str(v) for k, v in arg_dict.items()}, default=str)
    except Exception:
        return "Error capturing inputs"



import atexit
import os
from typing import Dict, Optional

from pydantic import BaseModel, Field, model_validator


class TelemetryBase(BaseModel):
    service_name: str = "gwenflow-service"
    protocol: str = "HTTP"
    endpoint: Optional[str] = None
    headers: Dict[str, str] = Field(default_factory=dict)

    @model_validator(mode='after')
    def set_default_endpoint(self) -> 'TelemetryBase':
        if not self.endpoint:
            env_endpoint = os.getenv("TELEMETRY_ENDPOINT")
            if env_endpoint:
                self.endpoint = env_endpoint
            else:
                if self.protocol.upper() == "HTTP":
                    self.endpoint = "http://localhost:4318"
                else:
                    self.endpoint = "localhost:4317" # For gRPC
        return self

    def initialize(self) -> None:
        try:
            from opentelemetry import trace
            from opentelemetry.sdk.resources import Resource
            from opentelemetry.sdk.trace import TracerProvider
            from opentelemetry.sdk.trace.export import BatchSpanProcessor
        except ImportError as exc:
            raise ImportError(
                "Opentelemetry is not installed.\n"
                "To enable it, install the required packages: "
                "`uv add opentelemetry-api opentelemetry-sdk opentelemetry-exporter-otlp openinference-semantic-conventions`"
            ) from exc

        current_provider = trace.get_tracer_provider()

        if not isinstance(current_provider, TracerProvider):

            resource = Resource.create({
                "service.name": self.service_name,
            })
            provider = TracerProvider(resource=resource)

            if self.protocol.upper() == "GRPC":
                from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter as GRPCOTLPExporter
                exporter = GRPCOTLPExporter(endpoint=self.endpoint, headers=self.headers)
            else:
                from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter as HTTPOTLPExporter
                exporter = HTTPOTLPExporter(endpoint=self.endpoint, headers=self.headers)

            processor = BatchSpanProcessor(exporter)
            provider.add_span_processor(processor)
            trace.set_tracer_provider(provider)
            atexit.register(provider.shutdown)



import uuid
from typing import Optional, Any, List, Dict
from typing_extensions import Literal
from pydantic import BaseModel, Field, field_validator, UUID4, ConfigDict
from time import time

from gwenflow.types import Message, Usage


class Function(BaseModel):
    name: str
    arguments: str

class ToolCall(BaseModel):
    id: str
    function: Function
    type: Literal["function"]

class ToolResponse(BaseModel):
    tool_call_id: str
    tool_name: str
    tool_args: Optional[Dict[str, Any]] = None
    tool_call_error: Optional[bool] = None
    result: Optional[str] = None
    usage: Optional[Usage] = None
    created_at: int = Field(default_factory=lambda: int(time()))

    def to_message(self) -> Message:
        return Message(
            role="tool",
            tool_call_id=self.tool_call_id,
            tool_name=self.tool_name,
            content=self.result,
        )

class ModelResponse(BaseModel):

    id: str = Field(default_factory=lambda: str(uuid.uuid4()))

    role: Optional[str] = None

    content: Optional[str] = None
    parsed: Optional[Any] = None
    reasoning_content: Optional[str] = None
    tool_calls: List[ToolCall] = Field(default_factory=list)

    finish_reason: Optional[str] = None
    usage: Optional[Usage] = None

    created_at: int = Field(default_factory=lambda: int(time()))

    def to_message(self) -> Message:
        return Message(**self.model_dump(exclude_unset=True))


class AgentResponse(BaseModel):

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    """The id of the response."""

    content: Optional[str] = None
    parsed: Optional[Any] = None
    reasoning_content: Optional[str] = None

    messages: list[Message] = Field(default_factory=list)
    """A list of messages generated by the agent."""

    finish_reason: Optional[str] = None
    usage: Usage = Field(default_factory=Usage)

    created_at: int = Field(default_factory=lambda: int(time()))

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("id", mode="before")
    @classmethod
    def deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise ValueError("This field is not to be set by the user.")

    def to_messages(self) -> list[Message]:
        """Convert the output into a list of input items suitable for passing to the model."""
        return [Message(**it.model_dump(exclude_unset=True)) for it in self.messages]  # type: ignore


from pydantic import Field

from gwenflow.readers.website import WebsiteReader
from gwenflow.tools import BaseTool


class WebsiteReaderTool(BaseTool):
    name: str = "WebsiteReaderTool"
    description: str = "Fetches and returns the content of a given URL."
    max_depth: int = 1

    def _run(self, url: str = Field(description="The url of the website to read.")):
        clean_documents = []
        for doc in WebsiteReader(max_depth=self.max_depth).read(url):
            clean_documents.append({"content": doc.content, "url": doc.metadata["url"]})
        return clean_documents



from typing import List, Literal, Optional, Union

from pydantic import Field, RootModel, model_validator

from gwenflow.types.responses.base import Logprob, ResponseBase


class ResponseOutputItemEvent(ResponseBase):
    type: Literal[
        'response.output_item.added',
        'response.output_item.done'
    ]
    item: str
    status: Literal['created', 'done']


class ResponseReasoningEvent(ResponseBase):
    type: Literal[
        'response.reasoning_summary_part.added',
        'response.reasoning_summary_part.done',
    ]
    item_id: str

class ResponseReasoningDeltaEvent(ResponseBase):
    type: Literal[
        'response.reasoning_summary_text.created',
        'response.reasoning_summary_text.delta',
        'response.reasoning_summary_text.done']
    delta: Optional[str] = None
    item_id: str


class ResponseToolCallEvent(ResponseBase):
    type: str
    item_id: str
    status: Optional[str] = None

    @model_validator(mode="after")
    def infer_status(self) -> "ResponseToolCallEvent":
        if "searching" in self.type:
            self.status = "searching"
        elif "completed" in self.type:
            self.status = "completed"
        return self

class ResponseContentEvent(ResponseBase):
    type: Literal[
        'response.content_part.added',
        'response.content_part.done']
    item_id: str
    logprobs: Optional[List[Logprob]] = Field(default_factory=list)
    text: Optional[str] = None

class ResponseContentDeltaEvent(ResponseBase):
    type: Literal[
    'response.output_text.created',
    'response.output_text.delta',
    'response.output_text.done']
    delta: Optional[str] = None
    item_id: str
    logprobs: Optional[List[Logprob]] = Field(default_factory=list)


AnyResponseEvent = Union[
        ResponseReasoningEvent,
        ResponseReasoningDeltaEvent,
        ResponseContentEvent,
        ResponseContentDeltaEvent,
        ResponseToolCallEvent,
    ]

class ResponseEvent(RootModel):
    root: AnyResponseEvent



import json
import os
from typing import Any, AsyncIterator, Dict, Iterator, List, Literal, Optional, Union

from openai import AsyncOpenAI, OpenAI
from pydantic import Field

from gwenflow.llms.base import ChatBase
from gwenflow.telemetry.base import TelemetryBase
from gwenflow.telemetry.openai.openai_instrument import openai_telemetry
from gwenflow.types import ItemHelpers, Message
from gwenflow.types.responses import (
    Response,
    ResponseContentDeltaEvent,
    ResponseContentEvent,
    ResponseEvent,
    ResponseReasoningDeltaEvent,
    ResponseReasoningEvent,
    ResponseToolCallEvent,
)
from gwenflow.utils import extract_json_str


class ResponseOpenAI(ChatBase):
    model: str = "gpt-5-mini"

    # model parameters
    background: Optional[bool] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    max_output_tokens: Optional[int] = None
    max_tool_calls: Optional[int] = None
    prompt_cache_key: Optional[bool] = None
    prompt_cache_retention: Optional[bool] = None
    text_format: Optional[Any] = None # TODO change later to available output type
    top_logprobs: Optional[int] = None
    reasoning_effort: Optional[Literal['low', 'medium', 'high']] = None
    reasoning_summary: Optional[Literal['auto', 'concise', 'detailed']] = None #Use only auto to make sure it is compatible with all reasoning models for now
    show_reasoning: bool = False

    # clients
    client: Optional[OpenAI] = None
    async_client: Optional[AsyncOpenAI] = None

    # client parameters
    api_key: Optional[str] = None
    organization: Optional[str] = None
    base_url: Optional[str] = None
    timeout: Optional[Union[float, int]] = None
    max_retries: Optional[int] = None

    # telemetry #TODO move this elsewhere
    service_name: str = Field(default="gwenflow-service")
    provider: Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        telemetry_config = TelemetryBase(service_name=self.service_name)
        self.provider = telemetry_config.setup_telemetry()

        openai_telemetry.instrument()

    def _get_client_params(self) -> Dict[str, Any]:
        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the OPENAI_API_KEY environment variable"
            )

        organization = self.organization
        if organization is None:
            organization = os.environ.get("OPENAI_ORG_ID")

        client_params = {
            "api_key": api_key,
            "organization": organization,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params

    @property
    def _model_params(self) -> Dict[str, Any]:

        model_params = {
            "background": self.background,
            "max_output_tokens": self.max_output_tokens,
            "max_tool_calls": self.max_tool_calls,
            "prompt_cache_key": self.prompt_cache_key,
            "prompt_cache_retention": self.prompt_cache_retention,
            "temperature": self.temperature,
            "texte.format": self.text_format,
            "top_logprobs": self.top_logprobs,
            "top_p": self.top_p,
        }
        if self.get_reasoning_model():
            model_params["reasoning"] = {
                "effort": self.reasoning_effort,
                "summary": self.reasoning_summary
            }

        if self.tools and self.tool_type == "base":
            model_params["tools"] = [tool.to_openai_new() for tool in self.tools]
            model_params["tool_choice"] = self.tool_choice or "auto"

        model_params = {k: v for k, v in model_params.items() if v is not None}

        return model_params

    def get_client(self) -> OpenAI:
        if self.client:
            return self.client
        client_params = self._get_client_params()
        self.client = OpenAI(**client_params)
        return self.client

    def get_async_client(self) -> AsyncOpenAI:
        if self.client:
            return self.client
        client_params = self._get_client_params()
        self.async_client = AsyncOpenAI(**client_params)
        return self.async_client

    def _parse_response(self, response: str, text_format: dict = None) -> str:
        """Process the response."""
        if text_format.get("type") == "json_object":
            try:
                json_str = extract_json_str(response)
                # text_response = dirtyjson.loads(json_str)
                text_response = json.loads(json_str)
                return text_response
            except Exception:
                pass

        return response

    def _format_message(self, message: Message) -> Dict[str, Any]:
        """Format a message into the format expected by OpenAI."""
        return message.to_openai()

    def invoke(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> Response:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            raw_response = self.get_client().responses.create(
                model=self.model,
                input=[self._format_message(m) for m in messages_for_model],
                **self._model_params,
            )

            response = Response.model_validate(raw_response.model_dump())
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}") from e

        if self.text_format:
            content_output = response.get_text()
            content = self._parse_response(
                content_output, text_format=self.text_format
            )
            for item in response.output:
                if item.type == "message" and item.content:
                    item.content[0].text = content
        return response

    async def ainvoke(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> Response:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            raw_response = await self.get_async_client().responses.create(
                model=self.model,
                input=[self._format_message(m) for m in messages_for_model],
                **self._model_params,
            )

            response = Response.model_validate(raw_response.model_dump())
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}") from e

        if self.text_format:
            content_output = response.get_text()
            content = self._parse_response(
                content_output, text_format=self.text_format
            )
            for item in response.output:
                if item.type == "message" and item.content:
                    item.content[0].text = content
        return response

    def stream(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> Iterator[ResponseEvent]:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            client = self.get_client()

            with client.responses.create(
                model=self.model,
                input=[self._format_message(m) for m in messages_for_model],
                stream=True,
                **self._model_params,
            ) as raw_stream:
                for raw_event in raw_stream:
                    try:
                        event_obj = ResponseEvent.model_validate(raw_event.model_dump())
                        event = event_obj.root

                        if isinstance(event, (ResponseReasoningEvent, ResponseReasoningDeltaEvent)):
                            if self.show_reasoning:
                                yield event_obj

                        elif isinstance(event, (ResponseContentEvent, ResponseContentDeltaEvent)):
                            yield event_obj

                        elif isinstance(event, ResponseToolCallEvent):
                            yield event_obj

                        if getattr(event, "type", None) == "response.done":
                            break

                    except Exception:
                        continue

        except Exception as e:
            raise RuntimeError(f"Erreur lors du stream OpenAI : {e}") from e

    async def astream(self, input: Union[str, List[Message]]) -> AsyncIterator[ResponseEvent]:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)

            client = self.get_async_client()

            async with await client.responses.create(
                model=self.model,
                input=[self._format_message(m) for m in messages_for_model],
                stream=True,
                **self._model_params,
                ) as raw_stream:
                async for raw_event in raw_stream:
                    try:
                        event_obj = ResponseEvent.model_validate(raw_event.model_dump())
                        event = event_obj.root

                        if isinstance(event, (ResponseReasoningEvent, ResponseReasoningDeltaEvent)):
                            if self.show_reasoning:
                                yield event_obj

                        elif isinstance(event, (ResponseContentEvent, ResponseContentDeltaEvent)):
                            yield event_obj

                        elif isinstance(event, ResponseToolCallEvent):
                            yield event_obj

                        if getattr(event, "type", None) == "response.done":
                            break

                    except Exception:
                        continue
        except Exception as e:
            raise RuntimeError(f"Erreur lors du stream OpenAI : {e}") from e



import json
import re
import uuid
from typing import Dict, Iterator, List, Optional, Union

from pydantic import BaseModel

from gwenflow.agents.agent import DEFAULT_MAX_TURNS, Agent
from gwenflow.logger import logger
from gwenflow.types import AgentResponse, ItemHelpers, Message, ToolCall, Usage

PROMPT_TOOLS = """\
You have access to the following tools. Only use these tools.

```json
{tools}
```
"""

PROMPT_REACT = """\
Please answer in the following format:

```
Thought: analyze the problem, plan the next action.
Action: tool name (one of {tool_names}) if using a tool.
Action Input: the input to the tool, in a JSON format representing the kwargs (e.g. {{"input": "hello world", "num_beams": 5}})
```

Please ALWAYS start with a Thought.

NEVER surround your response with markdown code markers. You may use code markers within your response if you need to.

Please use a valid JSON format for the Action Input. Do NOT do this {{'input': 'hello world', 'num_beams': 5}}.

If this format is used, the tool will respond in the following format:

```
Observation: tool response
```

You should keep repeating the above format till you have enough information to answer the question without using any more tools. At that point, you MUST respond in one of the following two formats:

```
Thought: I can answer without using any more tools. I'll use the user's language to answer
Final Answer: [your answer here (In the same language as the user's question)]
```

```
Thought: I cannot answer the question with the provided tools.
Final Answer: [your answer here (In the same language as the user's question)]
```

Question: {query}"""


SPECIAL_TOK_THOUGHT = "Thought:"
SPECIAL_TOK_ACTION = "Action:"
SPECIAL_TOK_FINAL_ANSWER = "Final Answer:"


class ReactAgentAction(BaseModel):
    thought: Optional[str] = None
    action: str
    action_input: str

    def get_tool_call(self) -> ToolCall:
        return ToolCall(
            id=str(uuid.uuid4()),
            function=self.action,
            arguments=json.loads(self.action_input),
        )


class ReactAgentFinish(BaseModel):
    final_answer: str


class ReactMessageParser(BaseModel):
    @classmethod
    def parse(cls, text: str) -> Union[ReactAgentAction, ReactAgentFinish]:
        includes_answer = SPECIAL_TOK_FINAL_ANSWER in text
        regex = r"Action\s*\d*\s*:[\s]*(.*?)[\s]*Action\s*\d*\s*Input\s*\d*\s*:[\s]*(.*)"
        action_match = re.search(regex, text, re.DOTALL)
        if action_match:
            if includes_answer:
                msg = f"Parsing LLM output produced both a final answer and a parse-able action: {text}"
                raise Exception(msg)
            action = action_match.group(1).strip()
            action_input = action_match.group(2)
            tool_input = action_input.strip(" ")
            tool_input = tool_input.strip('"')

            i = text.rfind(SPECIAL_TOK_ACTION)
            thought = text[:i].strip()
            if thought.startswith(SPECIAL_TOK_THOUGHT):
                thought = thought.split(SPECIAL_TOK_THOUGHT)[-1].strip()

            return ReactAgentAction(action=action, action_input=tool_input, thought=thought)

        if SPECIAL_TOK_FINAL_ANSWER in text:
            final_answer = text.split(SPECIAL_TOK_FINAL_ANSWER)[-1].strip()
            return ReactAgentFinish(final_answer=final_answer)

        msg = f"Could not parse: `{text}`"
        raise Exception(msg)


class ReactAgent(Agent):
    def _prepend_react_prompt(self, messages: List[Message]) -> List[Message]:
        tool_names = ",".join(t.name for t in self.tools)
        tool_descs = [t.to_openai()["function"] for t in self.tools]
        prompt_tools = PROMPT_TOOLS.format(tools=json.dumps(tool_descs, indent=2, ensure_ascii=False))

        messages[-1].content = (
            prompt_tools
            + "\n"
            + PROMPT_REACT.format(
                tool_names=tool_names,
                query=messages[-1].content,
            )
        )

        return messages

    def run(
        self,
        input: Union[str, List[Message], List[Dict[str, str]]],
        context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> AgentResponse:
        # prepare messages and task
        messages = ItemHelpers.input_to_message_list(input)
        task = messages[-1].content

        # add react prompt to the last message
        self.llm.tool_type = "react"
        messages = self._prepend_react_prompt(messages)

        # init agent response
        agent_response = AgentResponse()

        # history
        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        # add reasoning
        if self.reasoning_model:
            messages_for_reasoning_model = [m.to_dict() for m in self.history.get()]
            reasoning_agent_response = self.reason(messages_for_reasoning_model)
            usage = (
                Usage(
                    requests=1,
                    input_tokens=reasoning_agent_response.usage.input_tokens,
                    output_tokens=reasoning_agent_response.usage.output_tokens,
                    total_tokens=reasoning_agent_response.usage.total_tokens,
                )
                if reasoning_agent_response.usage
                else Usage()
            )
            agent_response.usage.add(usage)

        num_turns_available = DEFAULT_MAX_TURNS

        while num_turns_available > 0:
            num_turns_available -= 1

            # format messages
            messages_for_model = [m.to_dict() for m in self.history.get()]

            # call llm and tool
            response = self.llm.invoke(input=messages_for_model)

            # usage
            usage = (
                Usage(
                    requests=1,
                    input_tokens=response.usage.prompt_tokens,
                    output_tokens=response.usage.completion_tokens,
                    total_tokens=response.usage.total_tokens,
                )
                if response.usage
                else Usage()
            )
            agent_response.usage.add(usage)

            parsed_message = ReactMessageParser.parse(response.choices[0].message.content)

            # keep thought
            self.history.add_message(Message(role="assistant", content=response.choices[0].message.content))
            agent_response.messages.append(Message(role="assistant", content=response.choices[0].message.content))

            # stop if not tool call
            if isinstance(parsed_message, ReactAgentFinish):
                agent_response.content = parsed_message.final_answer
                agent_response.messages.append(Message(**response.choices[0].message.model_dump()))
                break

            # thinking
            agent_response.thinking = parsed_message.thought
            logger.debug(parsed_message.thought)

            # handle tool calls
            tool_message = self.run_tool(parsed_message.get_tool_call())
            self.history.add_message(Message(role="user", content=f"Observation: {tool_message.content}"))
            agent_response.messages.append(Message(role="user", content=f"Observation: {tool_message.content}"))

        # format response
        if self.response_model:
            agent_response.content = json.loads(agent_response.content)

        agent_response.finish_reason = "stop"

        return agent_response

    def run_stream(
        self,
        input: Union[str, List[Message], List[Dict[str, str]]],
        context: Optional[Union[str, Dict[str, str]]] = None,
    ) -> Iterator[AgentResponse]:
        # prepare messages and task
        messages = ItemHelpers.input_to_message_list(input)
        task = messages[-1].content

        # add react prompt to the last message
        self.llm.tool_type = "react"
        messages = self._prepend_react_prompt(messages)

        # init agent response
        agent_response = AgentResponse()

        # history
        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        # add reasoning
        if self.reasoning_model:
            messages_for_reasoning_model = [m.to_dict() for m in self.history.get()]
            reasoning_agent_response = self.reason(messages_for_reasoning_model)
            usage = (
                Usage(
                    requests=1,
                    input_tokens=reasoning_agent_response.usage.input_tokens,
                    output_tokens=reasoning_agent_response.usage.output_tokens,
                    total_tokens=reasoning_agent_response.usage.total_tokens,
                )
                if reasoning_agent_response.usage
                else Usage()
            )
            agent_response.usage.add(usage)

        num_turns_available = DEFAULT_MAX_TURNS

        while num_turns_available > 0:
            num_turns_available -= 1

            # format messages
            messages_for_model = [m.to_dict() for m in self.history.get()]

            # call llm and tool
            last_delta_message = None
            output = ""

            for chunk in self.llm.stream(input=messages_for_model):
                # usage
                usage = (
                    Usage(
                        requests=1,
                        input_tokens=chunk.usage.prompt_tokens,
                        output_tokens=chunk.usage.completion_tokens,
                        total_tokens=chunk.usage.total_tokens,
                    )
                    if chunk.usage
                    else Usage()
                )
                agent_response.usage.add(usage)

                if not chunk.choices or not chunk.choices[0].delta:
                    continue

                delta = chunk.choices[0].delta
                if not delta.content:
                    continue

                output += delta.content

                special_func_token = "Action: "
                special_args_token = "Action Input: "
                special_thought_token = "Thought: "
                special_final_token = "Final Answer: "

                a = output.rfind(special_func_token)  # action
                i = output.rfind(special_args_token)  # iaction nput
                t = output.rfind(special_thought_token)  # thought
                f = output.rfind(special_final_token)  # final

                last_delta_message = None
                if max([a, i, t, f]) == t:
                    agent_response.content = None
                    if last_delta_message != "thought":
                        last_delta_message = "thought"
                        agent_response.thinking = None
                    else:
                        agent_response.thinking = delta.content
                elif max([a, i, t, f]) == f:
                    agent_response.thinking = None
                    if last_delta_message != "final":
                        last_delta_message = "final"
                        agent_response.content = None
                    else:
                        agent_response.content = delta.content

                yield agent_response

            parsed_message = ReactMessageParser.parse(output)

            # keep thought
            self.history.add_message(Message(role="assistant", content=output))
            agent_response.messages.append(Message(role="assistant", content=output))

            # stop if no tool call
            if isinstance(parsed_message, ReactAgentFinish):
                agent_response.content = parsed_message.final_answer
                agent_response.messages.append(Message(role="assistant", content=output))
                break

            # thinking
            agent_response.thinking = parsed_message.thought
            if agent_response.thinking:
                logger.debug(parsed_message.thought)
                yield agent_response

            # handle tool calls
            observation = self.run_tool(parsed_message.get_tool_call())
            self.history.add_message(Message(role="user", content=f"Observation: {observation.content}"))
            agent_response.messages.append(Message(role="user", content=f"Observation: {observation.content}"))

        # format response
        if self.response_model:
            agent_response.content = json.loads(agent_response.content)

        agent_response.finish_reason = "stop"

        yield agent_response



import functools
import inspect
from contextlib import contextmanager
from typing import Any

from openinference.semconv.trace import OpenInferenceSpanKindValues, SpanAttributes
from opentelemetry import trace
from opentelemetry.trace import StatusCode

from .llm_utils import capture_llm_usage, capture_tool_calls, prepare_llm_attributes
from .utils import extract_user_inputs, safe_serialize


class DecoratorTracer:
    def __init__(self, tracer_name="gwenflow"):
        self.tracer = trace.get_tracer(tracer_name)

    @contextmanager
    def _start_span(self, name, kind, instance, func, args, kwargs):
        with self.tracer.start_as_current_span(name) as span:
            if span.is_recording():
                span.set_attribute(SpanAttributes.OPENINFERENCE_SPAN_KIND, kind.value)
                input_val = extract_user_inputs(func, (instance, *args), kwargs)
                span.set_attribute(SpanAttributes.INPUT_VALUE, str(input_val))

                if kind == OpenInferenceSpanKindValues.LLM:
                    prepare_llm_attributes(span, instance)

            try:
                yield span
            except Exception as e:
                if span.is_recording():
                    span.set_status(StatusCode.ERROR, str(e))
                    span.record_exception(e)
                raise

    def _finalize_span_output(self, span, content: str, tool_calls_json: str, fallback_obj: Any):
        if not span.is_recording():
            return

        if content:
            span.set_attribute(SpanAttributes.OUTPUT_VALUE, content)
        elif tool_calls_json:
            span.set_attribute(SpanAttributes.OUTPUT_VALUE, f"Tool Calls: {tool_calls_json}")
        elif fallback_obj is not None:
            span.set_attribute(SpanAttributes.OUTPUT_VALUE, safe_serialize(fallback_obj))

        span.set_status(StatusCode.OK)

    def _wrap_logic(self, name_attr, kind, name_override=None):
        def decorator(func):
            # 1. ASYNC GENERATOR (Streaming Async)
            if inspect.isasyncgenfunction(func):
                @functools.wraps(func)
                async def wrapper(instance, *args, **kwargs):
                    name = name_override or f"{kind.value}:{getattr(instance, name_attr, 'unknown')}"
                    with self._start_span(name, kind, instance, func, args, kwargs) as span:
                        last_chunk = None
                        accumulated_content = ""
                        latest_tool_calls = None

                        try:
                            async for chunk in func(instance, *args, **kwargs):
                                last_chunk = chunk

                                if hasattr(chunk, "content") and chunk.content:
                                    accumulated_content += chunk.content

                                if hasattr(chunk, "tool_calls") and chunk.tool_calls:
                                    latest_tool_calls = chunk.tool_calls

                                yield chunk

                            if kind == OpenInferenceSpanKindValues.LLM:
                                capture_llm_usage(span, last_chunk)
                            tc_json = capture_tool_calls(span, latest_tool_calls)
                            self._finalize_span_output(span, accumulated_content, tc_json, last_chunk)

                        except Exception as e:
                            if span.is_recording():
                                span.record_exception(e)
                                span.set_status(StatusCode.ERROR, str(e))
                            raise
                return wrapper

            # 2. ASYNC FUNCTION (ainvoke)
            elif inspect.iscoroutinefunction(func):
                @functools.wraps(func)
                async def wrapper(instance, *args, **kwargs):
                    name = name_override or f"{kind.value}:{getattr(instance, name_attr, 'unknown')}"
                    with self._start_span(name, kind, instance, func, args, kwargs) as span:
                        result = await func(instance, *args, **kwargs)

                        content = getattr(result, "content", "")
                        tc_json = None

                        if kind == OpenInferenceSpanKindValues.LLM:
                            capture_llm_usage(span, result)
                            tc_json = capture_tool_calls(span, getattr(result, "tool_calls", None))

                        self._finalize_span_output(span, content, tc_json, result)
                        return result
                return wrapper

            # 3. SYNC GENERATOR (Streaming Sync)
            elif inspect.isgeneratorfunction(func):
                @functools.wraps(func)
                def wrapper(instance, *args, **kwargs):
                    name = name_override or f"{kind.value}:{getattr(instance, name_attr, 'unknown')}"
                    with self._start_span(name, kind, instance, func, args, kwargs) as span:
                        last_chunk = None
                        accumulated_content = ""
                        latest_tool_calls = None

                        try:
                            for chunk in func(instance, *args, **kwargs):
                                last_chunk = chunk
                                if hasattr(chunk, "content") and chunk.content:
                                    accumulated_content += chunk.content
                                if hasattr(chunk, "tool_calls") and chunk.tool_calls:
                                    latest_tool_calls = chunk.tool_calls
                                yield chunk

                            if kind == OpenInferenceSpanKindValues.LLM:
                                capture_llm_usage(span, last_chunk)
                            tc_json = capture_tool_calls(span, latest_tool_calls)
                            self._finalize_span_output(span, accumulated_content, tc_json, last_chunk)

                        except Exception as e:
                            if span.is_recording():
                                span.record_exception(e)
                                span.set_status(StatusCode.ERROR, str(e))
                            raise
                return wrapper

            # 4. SYNC FUNCTION (invoke)
            else:
                @functools.wraps(func)
                def wrapper(instance, *args, **kwargs):
                    name = name_override or f"{kind.value}:{getattr(instance, name_attr, 'unknown')}"
                    with self._start_span(name, kind, instance, func, args, kwargs) as span:
                        result = func(instance, *args, **kwargs)

                        content = getattr(result, "content", "")
                        tc_json = None

                        if kind == OpenInferenceSpanKindValues.LLM:
                            capture_llm_usage(span, result)
                            tc_json = capture_tool_calls(span, getattr(result, "tool_calls", None))

                        self._finalize_span_output(span, content, tc_json, result)
                        return result
                return wrapper
        return decorator

    def llm(self, name=None):
        return self._wrap_logic("model", OpenInferenceSpanKindValues.LLM, name)

    def agent(self, name=None):
        return self._wrap_logic("name", OpenInferenceSpanKindValues.AGENT, name)

    def tool(self, name=None):
        return self._wrap_logic("name", OpenInferenceSpanKindValues.TOOL, name)

    def flow(self, name=None):
        return self._wrap_logic("name", OpenInferenceSpanKindValues.CHAIN, name)

tracer = DecoratorTracer()
